//#########################################################################
// File: LCD.C
//
// Ansteuerung mehrerer LCD-Anzeigen im
// 4 Bit Daten Mode
//
// Controller:
// HD44780, KS066, KS070 oder KS073. Siehe lcd.h.
//
// Todo:
// Noch einiges :(
//
// Die gnstigste Variante zur Beschaltung sieht so aus:
// Alle 4 Datenleitungen parallelschalten
// Alle RS Leitungen parallelschalten
// Nur die Enable Leitungen mssen jeweils an einen eigenen Pin
//
// Es ist aber auch mglich die LCDs vllig unabhngig
// voneinander zu verschalten.
//
// Drei LCDs an einem 8 Bit Port ist hiermit mglich.
// Oder 4x27 und 4x40 Displays mit zwei Controllern.
//
//#########################################################################
// Last change: 11.08.2010
//#########################################################################
// hk@holger-klabunde.de
// http://www.holger-klabunde.de/index.html
//#########################################################################
// Compiler: AVR-GCC 4.1.1
//#########################################################################
//@{
#include <avr/io.h>
#include <avr/pgmspace.h>

#if !defined (F_CPU)
 #define F_CPU  20000000 // je hher F_CPU desto lnger die Schleifen ;)
# warning "F_CPU not defined in lcd.c ! Assume 20.000000 MHz."
#endif

#include "lcd.h"

tLCD LCD1;		// struct for 1st LCD
tLCD LCD2;		// struct for 2nd LCD
tLCD LCD3;		// struct for 3rd LCD
				// you can do even more

void InitHD44780(tLCD *lcd);
void InitKS073(tLCD *lcd);
void InitKS066(tLCD *lcd);
void InitKS070(tLCD *lcd);

//*******************************************
// Set E high, wait a little bit and set E low
//*******************************************
void Clock_E(tLCD *lcd)
{
 *lcd->e_port |= lcd->e_mask;
 Delay1us(1);
 *lcd->e_port &= ~lcd->e_mask;
 Delay1us(1);
}

//*****************************************
// Commando ans Display senden
//*****************************************
void LCDCommand(tLCD *lcd, unsigned char command)
{
 *lcd->rs_port &= ~lcd->rs_mask; //Controlregister
 LCDWriteByte(lcd, command);
 *lcd->rs_port |= lcd->rs_mask;  //Zurck auf Datenregister
}

//*******************************************
// Schreibt ein Datennibble in die LCD-Anzeige
//*******************************************
void LCDWriteNibble(tLCD *lcd, unsigned char data)
{

 if(data & 0x01) {  *lcd->d4_port |= lcd->d4_mask; }
 else { *lcd->d4_port &= ~lcd->d4_mask; }

 if(data & 0x02) { *lcd->d5_port |= lcd->d5_mask; }
 else { *lcd->d5_port &= ~lcd->d5_mask; }

 if(data & 0x04) { *lcd->d6_port |= lcd->d6_mask; }
 else { *lcd->d6_port &= ~lcd->d6_mask; }

 if(data & 0x08) { *lcd->d7_port |= lcd->d7_mask; }
 else { *lcd->d7_port &= ~lcd->d7_mask; }

 Clock_E(lcd);
}

//*******************************************
// Schreibt ein Datenbyte in die LCD-Anzeige
//*******************************************
void LCDWriteByte(tLCD *lcd, unsigned char data)
{

 LCDWriteNibble(lcd, data >> 4);
 LCDWriteNibble(lcd, data & 0x0F);

 Delay1us(50); //etwas auf die Anzeige warten
}

//######################################################
//Schreibt einen Text aus dem RAM in die LCD-Anzeige
//######################################################
void LCDWrite(tLCD *lcd, char *s)
{
 char c;

  while((c = *s++))
   {
    c=ConvertChar(c);
    if(c) LCDWriteByte(lcd, c);
   }
}

//######################################################
// Schreibt einen Text aus dem FLASH in die LCD-Anzeige
// Aufruf LCDWrite_P(&LCD1,"text") ! Siehe lcd.h.
//######################################################
void LCDWrite_p(tLCD *lcd, const char *s)
{
 char c;

  while((c=pgm_read_byte(s++)))
   {
    c=ConvertChar(c);
    if(c) LCDWriteByte(lcd, c);
   }
}

//######################################################
// bersetzungstabelle fr Sonderzeichen
//######################################################
char ConvertChar(char data)
{
 char ps;

 ps = data;

#ifdef EADIP204
  //bersetzungstabelle Win zu EADIP204 Sonderzeichen
 switch(ps)
  {
   case '' : ps=0x7B; break; //Umlenkung auf neuen Zeichen Code
   case '' : ps=0x7C; break;
   case '' : ps=0x7E; break;
   case '' : ps=0xBE; break;
   case '' : ps=0x5B; break;
   case '' : ps=0x5C; break;
   case '' : ps=0x5E; break;
   case '' : ps=0x80; break;
   case '' : ps=0x8F; break;
   case '_' : ps=0xC4; break;
   case '@' : ps=0xA0; break;
   case '[' : ps=0xFA; break;
   case ']' : ps=0xFC; break;
   case '|' : ps=0xFE; break;
   case '{' : ps=0xFD; break;
   case '}' : ps=0xFF; break;
   case '~' : ps=0xCE; break;
   case '' : ps=0x5F; break;
   case '$' : ps=0xA2; break;
   case '^' : ps=0x1D; break;
   case '\\' : ps=0xFB; break;
   case 0x0A : ps=0; break;  //LF
   case 0x0D : ps=0; break;  //CR
  }
#elif defined CUSTOM_TABLE
  //bersetzungstabelle Sonderzeichen selbst definiert
 switch(ps)
  {
   case 0x0A : ps=0; break;  //LF
   case 0x0D : ps=0; break;  //CR
  }

#else
  //Standard bersetzungstabelle Win zu LCD Sonderzeichen
 switch(ps)
  {
   case '' : ps=0xE1; break; //Umlenkung auf neuen Zeichen Code
   case '' : ps=0xEF; break;
   case '' : ps=0xF5; break;
   case '' : ps=0xE2; break;
   case '' : ps=0xDF; break;
   case '' : ps=0xE4; break;
   case 0x0A : ps=0; break;  //LF
   case 0x0D : ps=0; break;  //CR
//   case '' : ps=0x00; break; //Umlenkung ins Character RAM
//   case '' : ps=0x01; break;
//   case '' : ps=0x02; break;
  }

#endif //#ifdef EADIP

 return ps;
}

//######################################################
// Springt auf eine  bestimmte Zeile und Spalte
// der LCD-Anzeige. Links oben ist Position 1,1
// Gltig fr bis zu 4 Zeilen
//######################################################
void LCDPos(tLCD *lcd, unsigned char zeile,unsigned char spalte)
{
 unsigned char buff;

 // Bei ungltigen Werten tue einfach nichts
 if(zeile == 0 || zeile > LCD_LINES) return;
 if(spalte == 0 || spalte > LCD_ROWS) return;

#if defined (HD44780) || defined (KS066) || defined (KS070)
// 0x80 wegen Display Adress Set.
// Offset's bei HD44780

 switch(zeile)
  {
   case 1: buff = 0x80 + 0x00 -1 + spalte; break;
#if LCD_LINES >1
   case 2: buff = 0x80 + 0x40 -1 + spalte; break;
#endif
#if LCD_LINES >2
 #if LCD_ROWS ==16
   case 3: buff = 0x80 + 0x10 -1 + spalte; break;
 #endif
 #if LCD_ROWS ==20
   case 3: buff = 0x80 + 0x14 -1 + spalte; break;
 #endif
#endif
#if LCD_LINES >3
 #if LCD_ROWS ==16
   case 4: buff = 0x80 + 0x50 -1 + spalte; break;
 #endif
 #if LCD_ROWS ==20
   case 4: buff = 0x80 + 0x54 -1 + spalte; break;
 #endif
#endif
   default : buff = 0x80; break; //1,1 remove compiler warning only
  }
#endif //#ifdef HD44780

#ifdef KS073
// Offset's anders als bei HD44780 Displays !
// 2 Lines
// Zeile 1 -> 0x80+0x00 = 0x80
// Zeile 2 -> 0x80+0x40 = 0xC0

// 4 Lines
// Zeile 1 -> 0x80+0x00 = 0x80
// Zeile 2 -> 0x80+0x20 = 0xA0
// Zeile 3 -> 0x80+0x40 = 0xC0
// Zeile 4 -> 0x80+0x60 = 0xE0

 switch(zeile)
  {
   case 1: buff = 0x80 + 0x00 -1 + spalte; break;
#if LCD_LINES ==2
   case 2: buff = 0x80 + 0x40 -1 + spalte; break;
#endif

#if LCD_LINES >2
   case 2: buff = 0x80 + 0x20 -1 + spalte; break;
   case 3: buff = 0x80 + 0x40 -1 + spalte; break;
#endif
#if LCD_LINES >3
   case 4: buff = 0x80 + 0x60 -1 + spalte; break;
#endif
   default : buff = 0x80; break; //1,1 remove compiler warning only
  }
#endif // #ifdef KS073

 LCDCommand(lcd,buff);
 Delay1us(50); //etwas auf die Anzeige warten
}

//######################################################
// Lscht die LCD Anzeige
// Cursor steht dann auf Position 1,Zeile 1
// Gilt auch fr mehrzeilige Anzeigen
//######################################################
void LCDCls(tLCD *lcd)
{
 LCDCommand(lcd, 0x01); //CLS
 Delay1ms(5);
}

//######################################################
//Stellt die Funktionsweise der LCD Anzeige ein
//######################################################
void LCDInit(tLCD *lcd)
{

 *lcd->rs_ddr |= lcd->rs_mask; //LCD_RS is output
 *lcd->e_ddr |= lcd->e_mask;   //LCD_E is output

 *lcd->rs_port &= ~lcd->rs_mask; //Control Register
 *lcd->e_port &= ~lcd->e_mask;   //LCD-Anzeige abwhlen

 *lcd->d4_ddr |= lcd->d4_mask; // Set pin to output
 *lcd->d5_ddr |= lcd->d5_mask; // Set pin to output
 *lcd->d6_ddr |= lcd->d6_mask; // Set pin to output
 *lcd->d7_ddr |= lcd->d7_mask; // Set pin to output

 Delay1ms(255); //Mglichen Power On Reset abwarten

#ifdef HD44780
 InitHD44780(lcd);
#endif

#ifdef KS066
 InitKS066(lcd);
#endif

#ifdef KS070
 InitKS070(lcd);
#endif

#ifdef KS073
 InitKS073(lcd);
#endif

 LCDCommand(lcd, 0x08); 	//Display off
 LCDCls(lcd);
 LCDCommand(lcd, 0x06); 	//Increment, Display Freeze

// LCDCommand(lcd, 0x0F); 	//Display on, Cursor on ,Cursor blink
// LCDCommand(lcd, 0x0E); 	//Display on, Cursor on ,Cursor no blink
 LCDCommand(lcd, 0x0C); 	//Display on, Cursor off ,Cursor no blink
 LCDCommand(lcd, 0x14); 	//Cursor Move,Shift off,Right Shift
}

//#######################################################
// Initialisation for HD44780
//#######################################################
#ifdef HD44780
void InitHD44780(tLCD *lcd)
{
 unsigned char i;

//Software Reset laut Hitachi Datenblatt
 for(i=0; i<3; i++)
  {
   LCDWriteNibble(lcd,0x03);
   Delay1ms(100); //Ein bichen warten auf die Anzeige
  }

 LCDWriteNibble(lcd,0x02);
 Delay1ms(5);
 #if LCD_LINES >1
  LCDWriteByte(lcd,0x28);    //4 Bit Operation,2 Line,5x7 Font
 #else
  LCDWriteByte(lcd,0x20);    //4 Bit Operation,1 Line,5x7 Font
 #endif
 Delay1ms(5);
}
#endif //#ifdef HD44780

//#######################################################
// Initialisation for KS073
//#######################################################
#ifdef KS073
void InitKS073(tLCD *lcd)
{
 LCDWriteNibble(lcd,0x02);  //4 Bit Operation
 Delay1ms(5);
 LCDWriteByte(lcd,0x2C);    //4 Bit Operation, RE=1
 Delay1ms(5);

#if LCD_LINES >2
 LCDWriteByte(lcd,0x09);    //5 Font, 4 Line
#else
 LCDWriteByte(lcd,0x08);    //5 Font, 1 or 2 Line
#endif
 Delay1ms(5);

 LCDWriteByte(lcd,0x28);    //4 Bit Operation, RE=0
 Delay1ms(5);
}
#endif //#ifdef KS073

//#######################################################
// Initialisation for KS066
//#######################################################
#ifdef KS066
void InitKS066(tLCD *lcd)
{
 LCDWriteNibble(lcd,0x02);  //4 Bit Operation
 LCDWriteNibble(lcd,0x02);  //4 Bit Operation
 #if LCD_LINES >1
  LCDWriteNibble(lcd,0x0C);   //2 Lines, display on
 #else
  LCDWriteNibble(lcd,0x04);   //1 Line, display on
 #endif

 Delay1ms(5);

}
#endif //#ifdef KS066

//#######################################################
// Initialisation for KS070
//#######################################################
#ifdef KS070
void InitKS070(tLCD *lcd)
{
 LCDWriteNibble(lcd,0x02);  //4 Bit Operation
 LCDWriteNibble(lcd,0x02);  //4 Bit Operation
 #if LCD_LINES >1
  LCDWriteNibble(lcd,0x08);   //2 Lines
 #else
  LCDWriteNibble(lcd,0x00);   //1 Line
 #endif

 Delay1ms(5);

}
#endif //#ifdef KS070

//#######################################################
// Setzt ein neues Zeichen (0-7) im CGRAM.
// "buf" ist das Array mit den 8 Datenbytes.
void LCDSetCGRAM(tLCD *lcd, unsigned char adr, unsigned char *buf)
//#######################################################
{
 unsigned char offset;

 if(adr>7) return;

 offset = adr<<3; //adr * 8;
 offset += 0x40;  //CGRam Flag setzen
 LCDCommand(lcd,offset);

 for(offset=0; offset<8; offset++) //8 Codes pro Zeichen bertragen
  {
   LCDWriteByte(lcd,*buf++);
  }

 LCDCommand(lcd,0x80); //DD_RAM Adresse 0 setzen, besser ist das
}

//#######################################################
//Zeigt ein Byte im HexCode (ohne 0x vorne) an
void LCDHexChar(tLCD *lcd, unsigned char by)
//#######################################################
{
 unsigned char buff;

 buff=by>>4; //Highnibble zuerst
 if(buff<10) buff+='0'; //ASCII Code erzeugen
 else buff+=0x37;        //Grobuchstaben
 LCDWriteByte(lcd,buff);

 buff=by&0x0f; //Danach das Lownibble
 if(buff<10) buff+='0'; //ASCII Code erzeugen
 else buff+=0x37;        //Grobuchstaben
 LCDWriteByte(lcd,buff);
}

//#######################################################
void LCDHexInt(tLCD *lcd, unsigned int l)
//#######################################################
{
 LCDHexChar(lcd,(unsigned char)(l>>8));
 LCDHexChar(lcd,(unsigned char)(l));
}

//#######################################################
void LCDHexLong(tLCD *lcd, unsigned long l)
//#######################################################
{
 LCDHexChar(lcd,(unsigned char)(l>>24));
 LCDHexChar(lcd,(unsigned char)(l>>16));
 LCDHexChar(lcd,(unsigned char)(l>>8));
 LCDHexChar(lcd,(unsigned char)(l));
}

//###################################################################################
//Nicht sonderlich genau !! Braucht aber util/delay.h nicht
void Delay1ms(unsigned int time)
//###################################################################################
{
 while(time--) Delay1us(1000);
}

//###################################################################################
//Nicht sonderlich genau !! Braucht aber util/delay.h nicht
void Delay1us(unsigned int time)
//###################################################################################
{
 while(time--)
  {
   //Ein CPU-Clock = 0.250us bei 4MHz
   //Ein CPU-Clock = 0.166us bei 6MHz
   //Ein CPU-Clock = 0.125us bei 8MHz
   //Ein CPU-Clock = 0.0904us bei 11.0592MHz
   //Ein CPU-Clock = 0.0625us bei 16MHz
   //Ein CPU-Clock = 0.0500us bei 20MHz

#if F_CPU >= 20000000
   NOP; NOP; NOP; NOP; NOP; NOP; NOP; NOP;
   NOP; NOP; NOP; NOP; NOP; NOP; NOP; NOP;
   NOP; NOP;
#elif F_CPU >= 16000000
   NOP; NOP; NOP; NOP; NOP; NOP; NOP; NOP;
   NOP; NOP; NOP; NOP; NOP; NOP;
#elif F_CPU >= 11059200
   NOP; NOP; NOP; NOP; NOP; NOP; NOP; NOP; NOP;
#elif F_CPU >= 8000000
    NOP; NOP; NOP; NOP; NOP; NOP;
#elif F_CPU >= 6000000
   NOP; NOP; NOP; NOP;
#elif F_CPU >= 4000000
   NOP; NOP;
#else
   NOP;
#endif

  }
}

//@}
